//=========================================================
//  MusE
//  Linux Music Editor
//    $Id: synth.h,v 1.22 2004/07/14 15:27:26 wschweer Exp $
//  (C) Copyright 2000-2004 Werner Schweer (ws@seh.de)
//=========================================================

#ifndef __SYNTH_H__
#define __SYNTH_H__

#include <qfileinfo.h>
#include <vector>

#include "globals.h"
#include "node.h"
#include "instruments/minstrument.h"
#include "mididev.h"
#include "midiport.h"
#include "track.h"

class MidiEvent;
class Mess;
struct MESS;
struct _FSTHandle;
struct _FST;

//---------------------------------------------------------
//   Synth
//    software synthesizer
//---------------------------------------------------------

class Synth {
      const MESS* descr;
      bool vstType;
      _FSTHandle* fstHandle;

      QFileInfo info;
      int _instances;

   public:
      Synth(const QFileInfo& fi, bool type);
      const char* description() const;
      const char* version()     const;

      int instances() const       { return _instances; }
      void incInstances(int val);
      QString baseName() const    { return info.baseName(); }
      QString path() const        { return info.dirPath(true); }
      void* instantiate();
      bool isVst() const          { return vstType; }
      };

class Mess;

//---------------------------------------------------------
//   SynthI
//    software synthesizer instance
//---------------------------------------------------------

class SynthI : public AudioTrack, public MidiDevice,
   public MidiInstrument
      {
      Synth* synthesizer;
      Mess* _mess;
      _FST* _fst;
      bool _guiVisible;
      std::vector<float> initParams;

      virtual bool getData(unsigned, int, unsigned, float**);
      virtual QString open()       { return QString("OK");}
      virtual void close()         {}
      virtual bool putEvent(const MidiPlayEvent&);
      virtual bool putMidiEvent(const MidiPlayEvent&) { return false;}
      virtual Track* newTrack() const { return 0; }

   public:
      SynthI();
      SynthI* clone() const { return new SynthI(*this); }
      ~SynthI();

      void read(Xml&);
      virtual void write(int, Xml&) const;

      void setName(const QString& s);
      QString name() const          { return AudioTrack::name(); }

      Synth* synth() const          { return synthesizer; }
      virtual bool isSynti() const  { return true; }
      const char* getPatchName(int, int, int) const;

      virtual const char* getPatchName(int, int, MType);
      virtual void populatePatchPopup(QPopupMenu*, int ch, MType);
      void setParameter(const char* name, const char* value) const;

      bool guiVisible() const;
      void showGui(bool v);
      bool hasGui() const;
      void getGeometry(int* x, int* y, int* w, int* h) const;
      void setGeometry(int x, int y, int w, int h);

      MidiPlayEvent receiveEvent();
      int eventsPending() const;

      bool init(Synth* s, const QString& instanceName);
      void deactivate2();
      void deactivate3();
      bool isActivated() const { return synthesizer && (_fst || _mess); }
      virtual bool hasAuxSend() const  { return true; }
      };

#include <vector>

extern std::vector<Synth*> synthis;  // array of available synthis
#endif

